#  jlibtls.tcl --
#  
#      This file is part of the jabberlib. It provides support for the
#      tls network socket security layer.
#      
#  Copyright (c) 2005 Sergei Golovan <sgolovan@nes.ru>
#  Based on jlibtls by Mats Bengtsson
#  
# $Id: jlibtls.tcl,v 1.1 2005/09/15 18:58:03 aleksey Exp $
#
# SYNOPSIS
#   jlibtls::new connid args
#	creates auth token
#	args: -certfile certfile
#	      -cafile   cafile
#	      -keyfile  keyfile
#	      -command  callback
#
#   token configure args
#	configures token parameters
#	args: the same as in jlibtls::new
#
#   token starttls args
#	starts STARTTLS procedure
#	args: the same as in jlibtls::new
#
#   token free
#	frees token resourses

##########################################################################

package require tls 1.4
package require namespaces 1.0

package provide jlibtls 1.0

##########################################################################

namespace eval jlibtls {
    variable uid 0
}

##########################################################################

proc jlibtls::new {connid args} {
    variable uid

    set token [namespace current]::[incr uid]
    variable $token
    upvar 0 $token state

    ::LOG "(jlibtls::new $connid) $token"

    set state(-connid) $connid
    catch { unset state(-starttls) }
    catch { unset state(-required) }

    proc $token {cmd args} \
	"eval {[namespace current]::\$cmd} {$token} \$args"

    eval [list $token configure] $args

    jlib::register_xmlns $state(-connid) $::NS(tls) [list $token parse]

    return $token
}

##########################################################################

proc jlibtls::free {token} {
    variable $token
    upvar 0 $token state

    ::LOG "(jlibtls::free $token)"

    jlib::unregister_xmlns $state(-connid) $::NS(tls)

    catch { unset state }
    catch { rename $token "" }
}

##########################################################################

proc jlibtls::configure {token args} {
    variable $token
    upvar 0 $token state

    ::LOG "(jlibtls::configure $token)"

    foreach {key val} $args {
	switch -- $key {
	    -certfile -
	    -cafile -
	    -keyfile -
	    -command {
		set state($key) $val
	    }
	    default {
		return -code error "Illegal option \"$key\""
	    }
	}
    }
}

##########################################################################

proc jlibtls::parse {token xmldata} {
    variable $token
    upvar 0 $token state

    jlib::wrapper:splitxml $xmldata tag vars isempty cdata children

    switch -- $tag {
	starttls {
	    set state(-starttls) 1
	    foreach ch $children {
		jlib::wrapper:splitxml $ch tag1 vars1 isempty1 cdata1 children1
		if {$tag1 == "required"} {
		    set state(-required) 1
		}
	    }
	}
	proceed {
	    $token proceed
	}
	failure {
	    $token failure $children
	}
    }
}

##########################################################################

proc jlibtls::starttls {token args} {
    variable $token
    upvar 0 $token state

    ::LOG "(jlibtls::start $token)"

    eval [list $token configure] $args

    jlib::trace_stream_features $state(-connid) [list $token tls_continue]
}

##########################################################################

proc jlibtls::tls_continue {token} {
    variable $token
    upvar 0 $token state
    
    ::LOG "(jlibtls::tls_continue $token)"

    if {![info exists state(-starttls)]} {
	set err [stanzaerror::error modify not-acceptable -text \
		     [::msgcat::mc \
			  "Server haven't provided STARTTLS feature"]]
	$token finish ERR [concat modify $err]
	return
    }

    set data [jlib::wrapper:createtag starttls -vars [list xmlns $::NS(tls)]]
    
    jlib::outmsg [jlib::wrapper:createxml $data] -connection $state(-connid)
}

##########################################################################

proc jlibtls::failure {token children} {
    variable $token
    upvar 0 $token state

    ::LOG "(jlibtls::failure $token)"

    set error [lindex $children 0]
    if {$error == ""} {
	set err [stanzaerror::error modify undefined-condition \
		     -text [::msgcat::mc "STARTTLS failed"]]
    } else {
	jlib::wrapper:splitxml $error tag vars empty cdata children
	set err [stanzaerror::error modify $tag]
    }
    $token finish ERR [concat modify $err]
}

##########################################################################

proc jlibtls::proceed {token} {
    variable $token
    upvar 0 $token state

    ::LOG "(jlibtls::proceed $token)"

    set args {}
    foreach key {-certfile -cafile -keyfile} {
	if {[info exists state($key)] && $state($key) != ""} {
	    lappend args $key $state($key)
	}
    }

    eval [list jlib::transport::tcp::to_tls $state(-connid)] $args

    jlib::reset $state(-connid)

    jlib::outmsg [jlib::wrapper:streamheader \
		      [jlib::connection_server $state(-connid)] \
		      -xml:lang [jlib::get_lang] -version "1.0"] \
	-connection $state(-connid)
    $token finish OK {}
}

##########################################################################

proc jlibtls::finish {token res xmldata} {
    variable $token
    upvar 0 $token state

    ::LOG "(jlibtls::finish $token) res"

    if {$res != "OK"} {
	jlib::client status [::msgcat::mc "STARTTLS failed"]
    } else {
	jlib::client status [::msgcat::mc "STARTTLS successful"]
    }
    uplevel #0 $state(-command) [list $res $xmldata]
}

##########################################################################

